using System.Collections;
using System;
using System.Collections.Generic;
using UnityEngine;
using UnityEngine.InputSystem;
using BigSock.Item;

namespace BigSock {
	// Common class for all characters.
	public partial class Character : Entity {

		/*
			Attack speed of the character.
		*/
		public float AttackSpeed => Stats.AttackSpeed;
		public float baseAttackSpeed = 1;

		/*
			Cooldown between attacks.
		*/
		public float AttackCooldown => 1.0f / AttackSpeed;
		
		/*
			Movement speed of the character.
		*/
		public float MovementSpeed => Stats.MoveSpeed;
		public float baseMovementSpeed = 1;

		
		/*
			Damage of the character.
		*/
		public float Damage => Stats.Damage;
		public float baseDamage = 1;

		/*
			Knockback force
		*/
		public float KnockbackForce => Stats.Knockback;
		public float knockbackForce = 150;

		
		/*
			Hit points of the character.
		*/
		public float HP { 
			get => baseHP; 
			set => baseHP = value; 
		}
		public float baseHP = 10;
	
		/*
			Mana of the character.
		*/
		public float Mana { get; set; }

		/*
			Stamina of the character.
		*/
		public float Stamina { get; set; }
		
		/*
			Maximum hit points of the character.
		*/
		public float MaxHP => Stats.MaxHP;
		public float baseMaxHP = 10;

		/*
			Xp....
		*/
		public float dropXP;
		public float xp;
		public float maxXp;
		public int level; 


		/*
			Indicates whether or not the characer is a live.
		*/
		public bool Alive { get; private set; } = true;
		
		/*
			Stores the next time the character can recieve damage.
		*/
		public DateTime NextTimeCanTakeDamage { get; private set; } = DateTime.Now;
		
		/*
			How long between each time the character can take damage.
		*/
		public TimeSpan IFrameDuration { 
			get => _iFrameDuration; 
			private set => _iFrameDuration = value; 
		}
		public TimeSpan _iFrameDuration = new TimeSpan(0, 0, 0, 0, 333);



		protected Rigidbody2D rb;


		/*
			The inventory of the character.
		*/
		public Inventory Inventory { get; protected set; }

		/*
			The base stats of the character.
		*/
		public ICharacterStats BaseStats { get; protected set; } = new CharacterStats();

		/*
			The final stats of the character after applying modifiers.
		*/
		public ICharacterStats Stats { get; protected set; } = new CharacterStats();



		protected virtual void Start() {
			rb = GetComponent<Rigidbody2D>();

			Inventory = new Inventory(this);

			// Set the base stats.
			BaseStats = new CharacterStats {
				MaxHP = baseMaxHP,
				MaxMana = 20,
				MaxStamina = 20,
				RegenMana = 2,
				RegenStamina = 2,
				
				Damage = baseDamage,
				MoveSpeed = baseMovementSpeed,
				Knockback = knockbackForce,
				Range = 1,
				AttackSpeed = baseAttackSpeed,
				CritChance = 1,
				CritDamageModifier = 1,
				ProjectileSpeed = 1,
			};
			
			UpdateModifiers();

			Mana = Stats.MaxMana;
			Stamina = Stats.MaxStamina;
		}

		/*
			Regenerates mana and stamina.
		*/
		protected virtual void Regenerate() {
			Mana = Math.Min(Stats.MaxMana, Mana + Time.fixedDeltaTime * Stats.RegenMana);
			Stamina = Math.Min(Stats.MaxStamina, Stamina + Time.fixedDeltaTime * Stats.RegenStamina);
		}

		/*
			Updates the modifiers to the character's stats.
		*/
		public virtual void UpdateModifiers(ICharacterStats modifiers = null) {
			modifiers ??= Inventory.Modifier;
			Stats = BaseStats.Modify(modifiers);
		}



		/*
			Add Kcockback. 
		*/
		//public void KnockBack(float force, Vector2 difference) {
		//	rb.AddForce(difference * force, ForceMode2D.Impulse);
		//}
		public void KnockBack(IAttackStats attack) {
			Vector2 difference = ((Vector2) transform.position - attack.Source).normalized;
			//KnockBack(attack.Knockback, difference);
			rb.AddForce(difference * attack.Knockback, ForceMode2D.Impulse);

		}

		/*
			Try to move in a given direction.
		*/
		protected virtual bool TryMove(Vector2 direction) {
			if(direction != Vector2.zero) {
				rb.AddForce(direction * (float) MovementSpeed * Time.fixedDeltaTime, ForceMode2D.Impulse);
				return true;
			}
			return false;
		}

		/*
			Adds damage to the character if they don't have IFrames.
		*/
		public virtual bool TakeDamage(AttackStats attack) {
			if(attack == null) throw new ArgumentNullException(nameof(attack));
			if(!attack.IsCalculated) throw new ArgumentException("Attack needs to be calculated.", nameof(attack));

			// Check if player has IFrames
			if(NextTimeCanTakeDamage > DateTime.Now)
				return false;

			// Start new IFrames
			NextTimeCanTakeDamage = DateTime.Now + IFrameDuration;

			// Trigger the event for taking damage.
			OnTakeDamage?.Invoke(this, attack.Actor, attack);

			// Inform the attacker they hit us.
			if(attack.Actor != null) attack.Actor.TargetHit(this, attack);
			

			// Add damage
			HP -= attack.Damage;
			AfterDamage(attack);

			TryKill(attack);

			return true;
		}


		/*
			Try to kill the character.
		*/
		public bool TryKill(AttackStats attack) {
			if(Alive && HP <= 0) {


				// Trigger the event for us dying.
				OnDeath?.Invoke(this, attack.Actor, attack);

				//== PUT CODE HERE TO HANDLE IF WE DODGED DEATH (In case we had an item to revieve or cheat death)

				Alive = false;

				// Inform the attacker killed us.
				if(attack.Actor != null) {
					attack.Actor.TargetKilled(this, attack);
					if(attack.Actor is PlayerController player)
						player.GainXp(dropXP);
				}

				AfterDeath();

				return true;
			}
			return false;
		}


		/*
			Method for healing the character.
		*/
		public bool TryHeal(float amount) {
			// Can't heal if full.
			if(HP >= MaxHP) {
				print($"[Character.TryHeal()] Already Full! ({HP:N1} >= {MaxHP:N1})");
				return false;
			}

			print($"[Character.TryHeal()] {HP:N1} + {amount:N1} = {HP + amount:N1}");
			OnHeal?.Invoke(this, amount);

			// Heal the character.
			var res = HP + amount;
			if(res > MaxHP) res = MaxHP;
			HP = res;
			return true;
		}

		/*
			Method for what to do when the character takes damage.
		*/
		protected virtual void AfterDamage(IAttackStats attack) {
			print($"[Character.AfterDamage()] {HP + attack.Damage:N1} - {attack.Damage:N1} = {HP:N1} {(attack.IsCrit ? "[CRIT]":"")}");
			KnockBack(attack);
		}

		/*
			Method for what to do when the character dies.
		*/
		protected virtual void AfterDeath() {
			print($"[Character.AfterDeath()] start. | {HP}, {Alive}");
			Destroy(gameObject);
		}
	}



	/*
		Items
	*/
	public partial class Character {
		
		/*
			Adds a listener for a conditional item to this character's event.
		*/
		public void AddItemListener(ConditionalItemBase item) {
			switch(item.Trigger) {
				case TriggerType.Fire: OnFire             += ((OnFireItemBase) item).Handler; break;
				case TriggerType.Hit: OnHit               += ((OnHitItemBase) item).Handler; break;
				case TriggerType.Kill: OnKill             += ((OnKillItemBase) item).Handler; break;

				case TriggerType.TakeDamage: OnTakeDamage += ((OnTakeDamageItemBase) item).Handler; break;
				case TriggerType.Heal: OnHeal             += ((OnHealItemBase) item).Handler; break;
				case TriggerType.Death: OnDeath           += ((OnDeathItemBase) item).Handler; break;

				default: print($"[Character.AddItemListener()] No handler for {item.Trigger}!"); break;
			}
		}

		/*
			Remove a listener for a conditional item from this character's event.
		*/
		public void RemoveItemListener(ConditionalItemBase item) {
			switch(item.Trigger) {
				case TriggerType.Fire: OnFire             -= ((OnFireItemBase) item).Handler; break;
				case TriggerType.Hit: OnHit               -= ((OnHitItemBase) item).Handler; break;
				case TriggerType.Kill: OnKill             -= ((OnKillItemBase) item).Handler; break;

				case TriggerType.TakeDamage: OnTakeDamage -= ((OnTakeDamageItemBase) item).Handler; break;
				case TriggerType.Heal: OnHeal             -= ((OnHealItemBase) item).Handler; break;
				case TriggerType.Death: OnDeath           -= ((OnDeathItemBase) item).Handler; break;

				default: print($"[Character.RemoveItemListener()] No handler for {item.Trigger}!"); break;
			}
		}

		/*
			Try to pick up an item.
		*/
		public bool TryPickUpItem(IItem item) {
			if(Inventory.AddItem(item)) {
				UpdateModifiers();
				print($"[Character.TryPickUpItem()] {item.Name} picked up. ({Inventory.Items.Count}/{Inventory.Cap})");
				return true;
			}
			print($"[Character.TryPickUpItem()] {item.Name} NOT picked up. ({Inventory.Items.Count}/{Inventory.Cap})");
			return false;
		}

		/*
			Try to drop an item.
		*/
		public bool TryDropItem(IItem item) {
			if(Inventory.RemoveItem(item)) {
				UpdateModifiers();
				print($"[Character.TryDropItem()] {item.Name} dropped. ({Inventory.Items.Count}/{Inventory.Cap})");
				return true;
			}
			print($"[Character.TryDropItem()] {item.Name} NOT dropped. ({Inventory.Items.Count}/{Inventory.Cap})");
			return false;
		}
	}

	/*
		Events
	*/
	public partial class Character {

		/*
			Trigers the OnHit event.
		*/
		public void TargetHit(Character target, AttackStats attack) {
			OnHit?.Invoke(this, target, attack);
		}

		/*
			Trigers the OnKill event.
		*/
		public void TargetKilled(Character target, AttackStats attack) {
			OnKill?.Invoke(this, target, attack);
		}
		

		public void GiveXp(float xp) {
			this.xp += xp; 
			OnGainedXP?.Invoke(this, xp);
		}

		/*
			Triggers when character uses an attack.
			Params: actor, attack stats.
		*/
		public event Action<Character, AttackStats> OnFire;

		/*
			Triggers when character hits an enemy.
			Params: actor, target, attack stats.
		*/
		public event Action<Character, Character, AttackStats> OnHit;

		/*
			Triggers when character kills an enemy.
			Params: actor, target, attack stats.
		*/
		public event Action<Character, Character, AttackStats> OnKill;



		/*
			Triggers when character takes damage.
			Params: actor, target, attack stats.
				(Target here is the character that dealt damage to this one)
		*/
		public event Action<Character, Character, AttackStats> OnTakeDamage;

		/*
			Triggers when character heals back health.
			Params: actor, amount.
				(Add heal source later on)
		*/
		public event Action<Character, float> OnHeal;

		/*
			Triggers when character has taken fatal damage. 
				(This is before the character is killed, so can be used to prevent death)
			Params: actor, target, attack stats.
				(Target here is the character that dealt damage to this one)
		*/
		public event Action<Character, Character, AttackStats> OnDeath;



		/*
			Triggers when character picked up an item.
			Params: actor, item.
		*/
		public event Action<Character, IItem> OnPickedUpItem;

		/*
			Triggers when character gains xp.
			Params: sorce, amount.
		*/
		public event Action<Character, float> OnGainedXP;

		/*
			Triggers when character levels up.
			Params: actor.
		*/
		public event Action<Character> OnLevelUp;

		

		/*
			Triggers when character enters a new room.
			Params: actor.
				(Add room object later)
		*/
		public event Action<Character> OnEnterNewRoom;

		/*
			Triggers when character clears a room.
			Params: actor.
				(Add room object later)
		*/
		public event Action<Character> OnClearedRoom;

		/*
			Triggers when character enters a new level.
			Params: actor.
				(Add level object later)
		*/
		public event Action<Character> OnEnterNewStage;

		
	}

}
